/*-
 * Copyright (C)2022 @BABOLO http://www.babolo.ru/
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ident "@(#) Copyright (C)2022 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: mod_authn_sqlite3.c,v 1.6 2022/12/11 04:05:15 babolo Exp $"

#include <apr_strings.h>
#include <apr_md5.h>
#include <ap_config.h>
#include <ap_provider.h>
#include <string.h>
#include <httpd.h>
#include <sqlite3.h>
#include <http_config.h>
#include <http_core.h>
#include <http_log.h>
#include <http_protocol.h>
#include <http_request.h>
#include <mod_auth.h>

#define MAXPWSZ 255
#define INDTEX  "mod_authn_sqlite3@BABOLO "
#define MER     APLOG_MARK, APLOG_ERR, 0, r, INDTEX
#define MCO     f, l, m, APLOG_ERR, 0, r, INDTEX
#define MT      "Auth_SQLite3_"

typedef struct {
    char      *dbfile;
    char      *passwd;
    char      *digest;
    char      *context;
    char      *fieldpass;
    char      *fieldlogin;
    char      *fieldcontext;
    char      *fieldcred;
    u_int32_t  busy_timeout;
    u_int16_t  numallpass;
    u_int16_t  maskallpass;
    char      *allpass[16];
    char       pop[];
} authn_sq_config_rec;

module AP_MODULE_DECLARE_DATA authn_sqlite3_module;

const static char f[] =
    "SELECT \"%s\".\"%s\"\n"
    " FROM \"%s\"\n"
    " WHERE \"%s\".\"%s\" = ? AND (\"%s\".\"%s\" = ? OR \"%s\".\"%s\" IS NULL)\n"
    "ORDER BY \"%s\".\"%s\"\n"
    "LIMIT 1\n"
    ";";

const static char c[] =
    "SELECT \"%s\".\"%s\"\n"
    " FROM \"%s\"\n"
    " WHERE \"%s\".\"%s\" = ? AND (\"%s\".\"%s\" = ? OR \"%s\".\"%s\" IS NULL) AND \"%s\".\"%s\" = ?\n"
    "ORDER BY \"%s\".\"%s\"\n"
    "LIMIT 1\n"
    ";";

const static char u[] =
    " *+,-./0123456789:=@ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz";

static const char *
setpass(cmd_parms *cmd, void *d, const char *ext) {
    authn_sq_config_rec *conf  = d;
    char                *erstr = NULL;
    size_t               s     = strlen(ext);

    if  (conf->numallpass > 15) {
        erstr = apr_psprintf(cmd->pool, "No more then 16 URIs");
    } else if (*ext != '/') {
        erstr = apr_psprintf(cmd->pool, "URI=%s~ but must begin with /", ext);
    } else if (!s) {
        erstr = apr_psprintf(cmd->pool, "Empty URI");
    } else {
        if  (ext[s - 1] == '*') {
            s--;
            conf->maskallpass |= 1 << conf->numallpass;
        }
        conf->allpass[conf->numallpass++] = strndup(ext, s);
    }
    return(erstr);
}

static const command_rec authn_sqlite3_cmds[] =
{   AP_INIT_TAKE1( MT "File"
                 , ap_set_file_slot
                 , (void *)APR_OFFSETOF(authn_sq_config_rec, dbfile)
                 , OR_AUTHCFG
                 , "SQLite3 DB file containing user IDs and passwords"
                 )
,   AP_INIT_TAKE1( MT "Passwd"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sq_config_rec, passwd)
                 , OR_AUTHCFG
                 , "Name of passwd table"
                 )
,   AP_INIT_TAKE1( MT "Digest"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sq_config_rec, digest)
                 , OR_AUTHCFG
                 , "Name of digest table"
                 )
,   AP_INIT_TAKE1( MT "Context"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sq_config_rec, context)
                 , OR_AUTHCFG
                 , "Context of AAA"
                 )
,   AP_INIT_TAKE1( MT "FieldLogin"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sq_config_rec, fieldlogin)
                 , OR_AUTHCFG
                 , "Name of login field"
                 )
,   AP_INIT_TAKE1( MT "FieldPass"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sq_config_rec, fieldpass)
                 , OR_AUTHCFG
                 , "Name of password field"
                 )
,   AP_INIT_TAKE1( MT "ContextField"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sq_config_rec, fieldcontext)
                 , OR_AUTHCFG
                 , "Name of context field"
                 )
,   AP_INIT_TAKE1( MT "CredField"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sq_config_rec, fieldcred)
                 , OR_AUTHCFG
                 , "Name of cred field"
                 )
,   AP_INIT_ITERATE( MT "AllPass"
                   , setpass
                   , NULL
                   , OR_AUTHCFG
                   , "All pass URI"
                   )
,   AP_INIT_TAKE1( MT "BusyTimeout"
                 , ap_set_int_slot
                 , (void *)APR_OFFSETOF(authn_sq_config_rec, busy_timeout)
                 , OR_AUTHCFG
                 , "Sqlite busy timeout"
                 )
,   {NULL}
};

#if 0
static void
testset(request_rec *r) {
    if  (!!r->connection->client_ip)
        apr_table_setn(r->subprocess_env, "TEST_CIP"  , r->connection->client_ip  );
    if  (!!r->connection->remote_host)
        apr_table_setn(r->subprocess_env, "TEST_RHOST", r->connection->remote_host);
    if  (!!r->connection->local_ip)
        apr_table_setn(r->subprocess_env, "TEST_LIP"  , r->connection->local_ip   );
    if  (!!r->connection->local_host)
        apr_table_setn(r->subprocess_env, "TEST_LHOST", r->connection->local_host );
    if  (!!r->uri)            apr_table_setn(r->subprocess_env, "TEST_URI"  , r->uri           );
    if  (!!r->args)           apr_table_setn(r->subprocess_env, "TEST_ARGS" , r->args          );
    if  (!!r->protocol)       apr_table_setn(r->subprocess_env, "TEST_PROTO", r->protocol      );
    if  (!!r->the_request)    apr_table_setn(r->subprocess_env, "TEST_REQ"  , r->the_request   );
    if  (!!r->hostname)       apr_table_setn(r->subprocess_env, "TEST_HOST" , r->hostname      );
    if  (!!r->method)         apr_table_setn(r->subprocess_env, "TEST_METH" , r->method        );
    if  (!!r->user)           apr_table_setn(r->subprocess_env, "TEST_USER" , r->user          );
    if  (!!r->unparsed_uri)   apr_table_setn(r->subprocess_env, "TEST_UURI" , r->unparsed_uri  );
    if  (!!r->path_info)      apr_table_setn(r->subprocess_env, "TEST_PATH" , r->path_info     );
    if  (!!r->useragent_ip)   apr_table_setn(r->subprocess_env, "TEST_UIP"  , r->useragent_ip  );
    if  (!!r->useragent_host) apr_table_setn(r->subprocess_env, "TEST_UHOST", r->useragent_host);
}
#endif

static void
compilerr( const char   *f
         , int           l
         , int           m
         , request_rec  *r
         , sqlite3      *odf
         , sqlite3_stmt *vm
         , const char   *c
         , int           ex
         , const char   *er
         , int           i
         ) {
    if  (!!ex) {
        if  (!i) {
            ap_log_rerror(MCO "%s %d: %s", er, ex, c);
        } else {
            ap_log_rerror(MCO "%s[%d] %d: %s", er, i, ex, c);
    }   }
    if  (!vm || !!(ex = sqlite3_finalize(vm))) {
        ap_log_rerror(MCO "Could not finalize %d", ex);
    }
    if  (!!(ex = sqlite3_close(odf))) {
        ap_log_rerror(MCO "Could not close %d", ex);
    }
    if  (!!(ex = sqlite3_shutdown())) {
        ap_log_rerror(MCO "Could not shutdown %d", ex);
}   }

static void *
create_authn_sqlite3_dir_config(apr_pool_t *p, char *d) {
    authn_sq_config_rec *conf;

    conf = apr_palloc(p, sizeof(authn_sq_config_rec) + strlen(c) + MAXPWSZ);
    bzero(conf, sizeof(authn_sq_config_rec) + strlen(c));
    return(conf);
}

static authn_status
check_password(request_rec *r, const char *user, const char *password) {
    authn_sq_config_rec *conf = ap_get_module_config(r->per_dir_config, &authn_sqlite3_module);
    sqlite3             *odf;
    sqlite3_stmt        *vm;
    int                  ex;
    int                  i;

    if  (!conf->dbfile) {
        ap_log_rerror(MER MT "File not specified in the config");
        return(AUTH_GENERAL_ERROR);
    }
    if  (!conf->passwd) {
        ap_log_rerror(MER MT "Passwd not specified in the config");
        return(AUTH_GENERAL_ERROR);
    }
    if  (!conf->fieldpass) conf->fieldpass = "passwd";
    if  (!conf->fieldlogin) conf->fieldlogin = "login";
    if  (!conf->fieldcontext) conf->fieldcontext = "context";
    if  (!conf->fieldcred) conf->fieldcred = "cred";
    if  (!conf->pop[0]) {
        size_t  sz = 0;

        if  (!!conf->passwd[strspn(conf->passwd, u)]) {
            ap_log_rerror(MER "Illegal table name: %s", conf->passwd);
            return(AUTH_GENERAL_ERROR);
        }
        if  (!!conf->fieldpass[strspn(conf->fieldpass, u)]) {
            ap_log_rerror(MER "Illegal password field name: %s", conf->fieldpass);
            return(AUTH_GENERAL_ERROR);
        }
        if  (!!conf->fieldlogin[strspn(conf->fieldlogin, u)]) {
            ap_log_rerror(MER "Illegal login field name: %s", conf->fieldlogin);
            return(AUTH_GENERAL_ERROR);
        }
        if  (!!conf->fieldcontext[strspn(conf->fieldcontext, u)]) {
            ap_log_rerror(MER "Illegal context field name: %s", conf->fieldcontext);
            return(AUTH_GENERAL_ERROR);
        }
        if  (!!conf->fieldcred[strspn(conf->fieldcred, u)]) {
            ap_log_rerror(MER "Illegal cred field name: %s", conf->fieldcred);
            return(AUTH_GENERAL_ERROR);
        }
        if  (!!conf->context) {
            sz = strlen(c) - 26
               + (7 * strlen(conf->passwd))
               + (2 * strlen(conf->fieldcred))
               + strlen(conf->fieldlogin)
               + (2 * strlen(conf->fieldpass))
               + strlen(conf->fieldcontext)
            ;
            if  (sz > strlen(c) + MAXPWSZ) {
                ap_log_rerror(MER "Too big DB names");
                return(AUTH_GENERAL_ERROR);
            }
            snprintf( conf->pop
                    , sz + 1
                    , c
                    , conf->passwd
                    , conf->fieldcred
                    , conf->passwd
                    , conf->passwd
                    , conf->fieldlogin
                    , conf->passwd
                    , conf->fieldpass
                    , conf->passwd
                    , conf->fieldpass
                    , conf->passwd
                    , conf->fieldcontext
                    , conf->passwd
                    , conf->fieldcred
                    );
        } else {
            sz = strlen(f) - 22
               + (6 * strlen(conf->passwd))
               + (2 * strlen(conf->fieldcred))
               + strlen(conf->fieldlogin)
               + (2 * strlen(conf->fieldpass))
            ;
            if  (sz > strlen(f) + MAXPWSZ) {
                ap_log_rerror(MER "Too big DB names");
                return(AUTH_GENERAL_ERROR);
            }
            snprintf( conf->pop
                    , sz + 1
                    , f
                    , conf->passwd
                    , conf->fieldcred
                    , conf->passwd
                    , conf->passwd
                    , conf->fieldlogin
                    , conf->passwd
                    , conf->fieldpass
                    , conf->passwd
                    , conf->fieldpass
                    , conf->passwd
                    , conf->fieldcred
                    );
    }   }
    if  (!!(ex = sqlite3_initialize())) {
        ap_log_rerror(MER "Could not initialize %d", ex);
        return(AUTH_GENERAL_ERROR);
    }
    if  (!!(ex = sqlite3_open_v2(conf->dbfile, &odf, SQLITE_OPEN_READONLY, NULL))) {
        sqlite3_shutdown();
        ap_log_rerror(MER "Could not open password DB %d: %s", ex, conf->dbfile);
        return(AUTH_GENERAL_ERROR);
    }
    if  (!conf->busy_timeout) {
        if  (!!(ex = sqlite3_busy_timeout(odf, conf->busy_timeout))) {
            compilerr(APLOG_MARK, r, odf, NULL, "", ex, "Fail busy_timeout", 0);
            return(AUTH_GENERAL_ERROR);
    }   }
    if  (!!(ex = sqlite3_prepare_v2(odf, conf->pop, -1, &vm, NULL))) {
        compilerr(APLOG_MARK, r, odf, vm, conf->pop, ex, "Could not compile", 0);
        return(AUTH_GENERAL_ERROR);
    }
    i = 0;
    if  (!!conf->context && !!(ex = sqlite3_bind_text(vm, ++i, conf->context, -1, SQLITE_STATIC))) {
        compilerr(APLOG_MARK, r, odf, vm, conf->context, ex, "Could not bind", i);
        return(AUTH_GENERAL_ERROR);
    }
    if  (!!(ex = sqlite3_bind_text(vm, ++i, user, -1, SQLITE_STATIC))) {
        compilerr(APLOG_MARK, r, odf, vm, user, ex, "Could not bind", i);
        return(AUTH_GENERAL_ERROR);
    }
    if  (!!(ex = sqlite3_bind_text(vm, ++i, password, -1, SQLITE_STATIC))) {
        compilerr(APLOG_MARK, r, odf, vm, password, ex, "Could not bind", i);
        return(AUTH_GENERAL_ERROR);
    }
    switch((ex = sqlite3_step(vm))) {
    case SQLITE_ROW: {
            const u_char *tuple = NULL;
            char         *cred;
            int           ncol;

            ncol = sqlite3_column_count(vm);
            if  (ncol != 1) {
                ap_log_rerror(MER "Col num %d IN %s", ncol, conf->pop);
                compilerr(APLOG_MARK, r, odf, vm, "", 0, "", 0);
                return(AUTH_GENERAL_ERROR);
            } else {
                if  (sqlite3_column_type(vm, 0) != SQLITE_NULL) tuple = sqlite3_column_text(vm, 0);
                if  (!!tuple) {
                    cred = apr_pstrdup(r->pool, (const char *)tuple);
                    apr_table_setn(r->subprocess_env, "CREDENTIALS", cred);
                }
/*******        testset(r); *******/
                compilerr(APLOG_MARK, r, odf, vm, "", 0, "", 0);
                return(AUTH_GRANTED);
        }   }
    case SQLITE_DONE:
        compilerr(APLOG_MARK, r, odf, vm, "", 0, "", 0);
        for (u_int16_t k = 0; k < conf->numallpass; ++k) {
            if  (conf->maskallpass & (1 << k)) {
                if  (!strncmp(r->uri, conf->allpass[k], strlen(conf->allpass[k]))) goto grant;
            } else {
                if  (!strcmp(r->uri, conf->allpass[k])) goto grant;
        }   }
        return(AUTH_DENIED);
grant:
        apr_table_setn(r->subprocess_env, "CREDENTIALS", "0");
        apr_table_setn(r->subprocess_env, "ALLTRYPASS", password);
/*******testset(r); *******/
        return(AUTH_GRANTED);
    }
    ap_log_rerror(MER "Could not SELECT step %d", ex);
    compilerr(APLOG_MARK, r, odf, vm, "", 0, "", 0);
    return(AUTH_GENERAL_ERROR);
}

static authn_status
get_realm_hash(request_rec *r, const char *user, const char *realm, char **rethash) {
    char                *sqlite3_hash = NULL;
    authn_sq_config_rec *conf = ap_get_module_config(r->per_dir_config, &authn_sqlite3_module);

    if  (!conf->dbfile) {
        ap_log_rerror(MER MT "File not specified in the config");
        return AUTH_GENERAL_ERROR;
    }




    if (!sqlite3_hash) return(AUTH_USER_NOT_FOUND);
    *rethash = sqlite3_hash;
    return(AUTH_USER_FOUND);
}

static const authn_provider authn_sqlite3_provider =
{   &check_password
,   &get_realm_hash
};

static void
register_hooks(apr_pool_t *p) {
    ap_register_auth_provider( p
                             , AUTHN_PROVIDER_GROUP
                             , "sqlite3"
                             , AUTHN_PROVIDER_VERSION
                             , &authn_sqlite3_provider
                             , AP_AUTH_INTERNAL_PER_CONF
                             );
}

AP_DECLARE_MODULE(authn_sqlite3) =
{   STANDARD20_MODULE_STUFF
,   create_authn_sqlite3_dir_config /* dir config creater */
,   NULL                            /* dir merger --- default is to override */
,   NULL                            /* server config */
,   NULL                            /* merge server config */
,   authn_sqlite3_cmds              /* command apr_table_t */
,   register_hooks                  /* register hooks */
};
